﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=10
@EndOfDesignText@
Sub Class_Globals
	Type production(Predecessor As String, Successor As String)
	Private fx As JFX
	Private Turtle As B4XTurtle
	Private Callback As Object
	Private p As Point

	Public Points As List
	Public Recording As List
	Public FocalPoint As Point
	Public StartPoint As Point
	Public closed As Boolean
End Sub

Public Sub Initialize(Callback_ As Object, Turtle_ As B4XTurtle) As SuperTurtle
	Callback = Callback_
	Turtle = Turtle_
	Points.Initialize
	Recording.Initialize
	Return Me
End Sub

Public Sub Position(pt As Point, Angle As Int)
	Turtle.PenUp.MoveTo(pt.X, pt.Y).SetAngle(Angle)
End Sub

Public Sub CircleAt(pt As Point, radius As Float)
	Dim stepSize As Float = ((radius*2)*cPI)/360
	Turtle.PenUp.MoveTo(pt.X + radius, pt.Y).PenDown
	Turtle.SetAngle(-90)
	For i = 1 To 360
		Turtle.MoveForward(stepSize).TurnLeft(1)
	Next
	Turtle.PenUp.MoveTo(20, Turtle.Height - 20).setAngle(0)
End Sub

Public Sub CrossHairsAt(pt As Point, radius As Float)
	Turtle.PenUp.SetPenSize(1)
	Turtle.MoveTo(pt.X, pt.Y - radius).PenDown.MoveTo(pt.X, pt.Y + radius).PenUp.MoveTo(pt.X - radius, pt.Y).PenDown.moveTo(pt.X + radius, pt.Y)
	Turtle.SetAngle(-90)
	Dim stepSize As Float = ((radius*2)*cPI)/360
	For i = 1 To 360
		Turtle.MoveForward(stepSize).TurnLeft(1)
	Next
	Turtle.PenUp.MoveTo(20, Turtle.Height - 20).setAngle(0)
End Sub

Public Sub HandleTouch(pt As Point)
	Turtle.ClearScreen
	Dim collision As Boolean
	For i = 0 To Points.Size - 1
		Dim pq As Point = Points.Get(i)
		If pq.Distance(pt) < 20 Then
			collision = True
			Exit
		End If
		CircleAt(pq, 5)
	Next
	If collision Then
		Points.RemoveAt(i)
	Else
		Points.Add(pt)
		CrossHairsAt(pt, 10)
	End If
End Sub

Public Sub RectangleAt(p1 As Point, p2 As Point)
	Turtle.PenUp.MoveTo(p1.X, p1.Y).PenDown
	Turtle.MoveTo(p2.X, p1.Y).MoveTo(p2.X, p2.Y).MoveTo(p1.X, p2.Y).MoveTo(p1.X, p1.Y)
	Turtle.PenUp.RabbitMode.MoveTo(20, Turtle.Height - 20).setAngle(0)
End Sub

Public Sub DotsAt(pts As List)
	For Each pt As Point In pts
		CircleAt(pt, 2)
	Next
End Sub

Public Sub CurveFor(pts As List, closeShape As Boolean)
	Select pts.Size
		Case 0, 1 'do nothing, no curve
		Case 2, 3, 4
			Bezier(pts)
		Case Else
			Smooth(pts, closeShape)
	End Select
End Sub

Private Sub Bezier(pts As List)
	Dim startAt As Point = pts.Get(0)
	Turtle.PenUp.MoveTo(startAt.X, startAt.Y).PenDown
	If pts.Size = 2 Then
		Dim target As Point = pts.Get(1)
		Turtle.MoveTo(target.X, target.Y)
	Else
		For t50 = 1 To 50
			Dim t As Float = t50 / 50
			Dim target As Point = Quadratic(t, pts.Get(0), pts.Get(1), pts.Get(2))
			If pts.Size = 4 Then target = target.MultiplyValue(1 - t).Add(Quadratic(t, pts.Get(1), pts.Get(2), pts.Get(3)).MultiplyValue(t))
			Turtle.MoveTo(target.X, target.Y)
		Next
	End If
	Turtle.PenUp.MoveTo(20, Turtle.Height - 20).setAngle(0)
End Sub

Private Sub Quadratic(t As Float, p0 As Point, p1 As Point, p2 As Point) As Point
	Dim beta0 As Float = (1 - t) * (1 - t)
	Dim beta1 As Float = t * t
	Return p1.Add(p0.Subtract(p1).MultiplyValue(beta0)).Add(p2.Subtract(p1).MultiplyValue(beta1))
End Sub

Private Sub Smooth(pts As List, closeShape As Boolean)
	Dim startAt As Point = pts.Get(0)
	Turtle.PenUp.MoveTo(startAt.X, startAt.Y).PenDown
	Dim smoothPath As List = ComputeSmooth(pts, 4, 4, closeShape)
	For i = 1 To smoothPath.Size - 1
		Dim pt As Point = smoothPath.Get(i)
		Turtle.MoveTo(pt.X, pt.Y)
	Next
	Turtle.PenUp.RabbitMode.MoveTo(20, Turtle.Height - 20).setAngle(0)
End Sub

Public Sub PointList(pts() As Object) As List
	Dim result As List
	result.Initialize
	For Each pt As Point In pts
		result.Add(pt)
	Next
	Return result
End Sub

Private Sub PointListCopy(a As List) As List
	Dim result As List: result.Initialize
	For Each pt As Point In a
		result.Add(p.XY(pt.x, pt.y))
	Next
	Return result
End Sub

Public Sub ComputeSmooth(OrigPoints As List, smoothness As Int, roundness As Int, closeShape As Boolean) As List
	Dim Points As List = PointListCopy(OrigPoints)
	If closeShape Then Points.Add(Points.Get(0).As(Point).Copy)
	Dim roundnessFactor As Float = .3 * roundness / 5
	Dim closed As Boolean = False
	Dim p0, p1, p2 As Point
	For m = 0 To (smoothness - 1)
		Dim p As Point = Points.Get(0)
		If closed Then Points.add(p.Copy)
		p0 = Points.Get(0)
		p1 = Points.Get(1)
		Dim k As Int = 2
		Do While k < Points.Size
			p2 = Points.Get(k)
			Points.Set(k-1, p0.PartwayTo(p1, 1 - roundnessFactor))
			Points.insertAt(k, p1.PartwayTo(p2, roundnessFactor))
			p0 = p1.copy
			p1 = p2.copy
			k = k + 2
		Loop
		If closed Then
			Points.RemoveAt(0)
			Dim p As Point = Points.Get(0)
			Points.Set(Points.Size - 1, p.Copy)
		End If
	Next
	Return Points
End Sub

Public Sub CollectPoints
	Turtle.SetTurtleVisible(False)
	For Each pt As Point In Points
		CircleAt(pt, 5)
	Next
	Wait For Collected_Points
	Turtle.SetTurtleVisible(True)
	CallSubDelayed(Callback, "Collection_Done")
End Sub

Public Sub Instructions(stepSize As Int, iterations As Int, angle As Int, initialValue As String, rules As Map)
	Dim productions As List
	productions.initialize
	Dim vocabulary As StringBuilder
	vocabulary.Initialize
	For Each predecessor As String 	In rules.Keys
		vocabulary.Append(predecessor)
		productions.Add(NewProduction(predecessor, rules.Get(predecessor)))
	Next
	Dim symbols As String = vocabulary.toString
	Sleep(20)
	FollowInstructions(entelechy(symbols, initialValue, productions, iterations), stepSize, angle)
End Sub

Private Sub entelechy(symbols As String, initialValue As String, productions As List, iterations As Int) As String
	Dim currentState As String = initialValue
	For i = 1 To iterations - 1
		Dim sb As StringBuilder
		sb.Initialize
		For j = 0 To currentState.Length - 1
			Dim c As String = currentState.CharAt(j)
			If symbols.contains(c) Then
				For Each r As production In productions
					If c = r.predecessor Then
						sb.Append(r.successor)
						Exit
					End If
				Next
			Else
				sb.Append(c)
			End If
		Next
		currentState = sb.toString
	Next
	
	Return sb.toString
End Sub

Private Sub FollowInstructions(Sequence As String, delta As Int, angle As Int)
	StartPoint = p.XY(-1, -1)
	Recording.Clear
	Dim reverse As Boolean
	Dim n As Int = Sequence.length
	For i = 0 To n - 1
		Dim c As String = Sequence.CharAt(i)
		Select c
			Case " ": Continue
			Case "A", "B", "F", "X": Turtle.PenDown.moveforward(delta)
			Case "a", "b", "f", "x": Turtle.PenUp.moveforward(delta)
			Case "|": reverse = Not(reverse)
			Case "+"
				If reverse Then Turtle.TurnRight(angle) Else Turtle.TurnLeft(angle)
			Case "-"
				If reverse Then Turtle.TurnLeft(angle) Else Turtle.TurnRight(angle)
			Case "^": Turtle.SetAngle(-90)	'vertical up
			Case "v": Turtle.SetAngle(90)	'vertical down
			Case ">": Turtle.SetAngle(0)	'horizontal right
			Case "<": Turtle.SetAngle(180)	'horizontal left
			Case "[": Turtle.PushState
			Case "]": Turtle.PopState
			Case "*"
				Dim pt As Point = p.XY(Turtle.GetX, Turtle.GetY)
				Recording.Add(pt)
				If StartPoint.X < 0 Then StartPoint = pt.copy
			Case ".": FocalPoint =p.XY(Turtle.GetX, Turtle.GetY)
			Case "{": Turtle.StartPolygon
			Case "}": Turtle.FillPolygon
			Case "#"
			Case "?"
				Turtle.SetPenColor(Turtle.RandomColor)
		End Select
	Next
	Turtle.PenUp
	Turtle.MoveTo(20, Turtle.Height - 20)
	CallSubDelayed(Callback, "Recording_Done")
End Sub

Public Sub SortRecording(indices() As Object) As List
	Dim result As List
	result.initialize
	For i = 0 To indices.Length - 1
		Dim pt As Point = Recording.Get(indices(i)).As(Point).copy
		result.Add(pt)
	Next
	Return result
End Sub

Public Sub ShowIndices(pnts As List)
	For i = 0 To pnts.Size - 1
		Dim p0 As Point = pnts.Get(i)
		Turtle.MoveTo(p0.X, p0.Y)
		Turtle.DrawText(i)
	Next
End Sub

Public Sub Scallops(pts As List, delta As Int, fill As Boolean, clr As Int)
	Dim saveColor As Int = Turtle.GetPenColor
	If fill Then Turtle.StartPolygon
	For i = 1 To pts.Size - 1
		Dim p0 As Point = pts.Get(i- 1)
		Dim p1 As Point = pts.Get(i)
		Dim pt As Point = p0.PartwayTo(p1, .5)
		Dim perp As Point = p.XY(p0.Y - p1.Y, p1.X - p0.X)
		Dim concave As Float = delta / p0.Distance(p1)
		Dim pza As Point = pt.Add(perp.MultiplyValue(concave))
		Dim pzb As Point = pt.Add(perp.MultiplyValue(-concave))
		Dim aDist As Float = pza.Distance(FocalPoint)
		Dim bDist As Float = pzb.Distance(FocalPoint)
		If Abs(aDist - bDist) < .001 Then
			CurveFor(PointList(Array(p0, pzb, p1)), False)
		Else
			CurveFor(PointList(Array(p0, pza, p1)), False)
		End If
	Next
	If fill Then
		Turtle.SetPenColor(clr)
		Turtle.fillPolygon
	End If
	Turtle.SetPenColor(saveColor)
	CallSubDelayed(Callback, "Scallops_Done")
End Sub

Public Sub NewProduction (Predecessor As String, Successor As String) As production
	Dim t1 As production
	t1.Initialize
	t1.Predecessor = Predecessor
	t1.Successor = Successor
	Return t1
End Sub